"""Package configurations."""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from datetime import timezone


AWS_IP_ADDRESS_RANGES_URL = "https://ip-ranges.amazonaws.com/ip-ranges.json"

CREATE_DATE_FORMAT = "%Y-%m-%d-%H-%M-%S"
CREATE_DATE_TIMEZONE = timezone.utc
